import os
from os import F_OK, R_OK, W_OK, X_OK
from os.path import join, dirname, isabs, basename, realpath, lexists, islink
import sys
from stat import S_IFDIR
import shutil
import subprocess
import grp
import resource

from .util import Checker, mount, make_dirs, get_fds

USAGE = """

NAME

  rainbow-inject

SYNOPSIS

  rainbow-inject is a security-wrapper for Sugar Activities. Its purpose is to
  launch them in a low-privilege environment so that, if they are exploited,
  they are able to do less harm than would be the case if they ran with the
  full authority of the XO's human operator.

  This isolation is accomplished by running each activity as a separate user
  from a set of users that is known, at system installation time, to have
  tightly restricted access to the file-system.

  The environment that rainbow-inject creates for each activity it starts is
  supposed to be controlled by an instance-specific security-record (usually
  derived from a prototype record stored per-bundle) but this is not yet
  implemented. Therefore, Rainbow currently offers no way to customize what
  kinds of isolation it provides.

ENVIRONMENT VARIABLES

  SUGAR_BUNDLE_PATH    : MUST contain an absolute path to a bundle to launch.

  SUGAR_BUNDLE_ID      : MUST contain the unique bundle-id of the bundle at
                         SUGAR_BUNDLE_PATH.

  RAINBOW_DIR          : Optional; defaults to $HOME/isolation/1
                         This directory is used by rainbow-inject for all its
                         run-time data: currently, this means uid/gid
                         allocation data and for storing persistent
                         activity-data storage

  RAINBOW_STRACE_LOG   : Optional; defaults to Nothing.
                         If set, rainbow-inject will cause strace to write a
                         complete log to stdout.

  RAINBOW_CONSTANT_UID : Optional; defaults to Nothing.
                         If set, rainbow-inject will cause strace to write a
                         complete log to stdout.

ARGUMENTS

  Any arguments will be exec'ed verbatim after rainbow-inject properly
  configures its environment.

FILE DESCRIPTORS

  rainbow-inject will close all open file descriptors except those marked
  'safe' by its caller.

  rainbow-inject will print trace information on descriptor 1 and may print
  error messages on descriptor 2.

RETURN CODES

  Not specified.

"""

# Locations of these two files come from the XAUTHORITY and ICEAUTHORITY
# environment variables which are, alas, not set yet when rainbow starts.
# So they're hard-coded here; values should match those set in
# olpc-dm (in olpc-utils package)
_AUTH_HOME='/var/tmp/olpc-auth'
_XAUTHORITY=_AUTH_HOME + '/.Xauthority'
_ICEAUTHORITY=_AUTH_HOME + '/.ICEauthority'

def strace(log, argv, env):
    # cjb thinks we should really be using SystemTap here and refers us to
    # wiki:SystemTap and #83 <MS>
    log('applying strace')
    args = ['/usr/bin/strace', '-f']
    for k, v in env.iteritems():
        args.append('-E')
        args.append('%s=%s' % (k, v))
    args.extend(argv)

    return args, {}

def reserve_elt(pool_dir, elt, max_elt, incr, elt_name):
    fd = None
    while elt < max_elt:
        try:
            fd = os.open(join(pool_dir, str(elt)), os.O_EXCL | os.O_CREAT, 0600)
            os.close(fd)
            break
        except OSError:
            elt += incr
    if fd is None:
        raise RuntimeError("No " + elt_name + " available.")
    return elt

def reserve_credentials(log, spool, bundle_id, constant_uid):
    # Pick a gid for yourself.
    gid = reserve_elt(join(spool, 'gid_pool'), 10001, 65534, 2, 'gids')

    # Then try to atomically symlink the gid you picked to 'name' in gid_dir.
    # If you succeed, then you have the right gid.
    # If you fail, someone else has the right gid so release yours and use theirs.
    name_path = join(spool, 'bundle_id_to_gid', bundle_id)
    gid_path = join(spool, 'gid_pool', str(gid))
    try:
        os.symlink(gid_path, name_path)
    except OSError:
        os.unlink(gid_path)
        gid = int(basename(realpath(name_path)))

    # Decide whether to reserve a new (even) uid or to set uid <- gid and reserve that.
    uid = constant_uid and gid or reserve_elt(join(spool, 'uid_pool'), 10000, 65534, 2, 'uids')
    open(join(spool, 'uid_pool', str(gid)), 'w').close()
    log('reserved credentials (%d, %d)', uid, gid)

    path = join(spool, 'uid_to_gid', str(uid))
    if not lexists(path):
        os.symlink(str(gid), path)
    else:
        assert islink(path) and os.readlink(path) == str(gid)

    return (uid, gid)

def grab_home(log, spool, bundle_id, uid, gid, owner_gid):
    home = join(spool, 'uid_to_home_dir', str(uid))
    comment = '%s.%d' % (bundle_id, uid)

    gid_cmd = ['/usr/sbin/groupadd', '-o', '-g', str(gid), str(gid)]
    log('adding group: %s', ' '.join(gid_cmd))
    try:
        subprocess.check_call(gid_cmd)
    except subprocess.CalledProcessError, e:
        if e.returncode != 9: # we don't care if the group already exists
            raise e

    uid_cmd = ['/usr/sbin/useradd', '-m', '-u', str(uid), '-g', str(gid),
                                    '-c', comment, '-d', home, str(uid)]
    log('adding user: %s', ' '.join(uid_cmd))
    try:
        subprocess.check_call(uid_cmd)
    except subprocess.CalledProcessError, e:
        if e.returncode == 9: # we don't care if the user already exists
            log('notice: home dir %s already exists', home)
        else:
            raise e

    os.chown(home, uid, owner_gid)
    os.chmod(home, 0770)

    return home

def configure_home(spool, owner_uid, owner_gid, uid, gid, home):
    # Beware of CONSTANT_UID + instance==home: the instance dir may already exist.
    path = join(spool, 'uid_to_instance_dir', str(uid))
    if not lexists(path):
        os.symlink(home, path)

    if not lexists(join(home, 'instance')):
        os.symlink('.', join(home, 'instance'))

    path = join(spool, 'gid_to_data_dir', str(gid))
    make_dirs(path, owner_uid, gid, 0770)
    os.chown(path, owner_uid, gid)
    os.chmod(path, 0770)
    if not lexists(join(home, 'data')):
        os.symlink(path, join(home, 'data'))

    make_dirs(join(home, 'tmp'), uid, gid, 0700)
    mount('tmpfs', join(home, 'tmp'), 'tmpfs', 0, '')
    # home is 770 <uid>.<owner_gid> and, in the normal implementation, is in a
    # separate namespace.

    xauth = join(home, '.Xauthority')
    shutil.copy(_XAUTHORITY, xauth)
    os.chown(xauth, uid, gid)

    iceauth = join(home, '.ICEauthority')
    shutil.copy(_ICEAUTHORITY, iceauth)
    os.chown(iceauth, uid, gid)

    for frag in ['session.info', 'config', 'owner.key.pub']:
        path = join('.sugar/default/', frag)
        make_dirs(dirname(join(home, path)), uid, owner_gid, 0770)
        shutil.copy(join('/home/olpc', path),
                    join(home, path))
        os.chown(join(home, path), uid, owner_gid)
        os.chmod(join(home, path), 0770)

def launch(log, home, uid, gid, argv, env, cwd, pset, safe_fds, strace_hint,
           preloader_hint):
    env['USER'] = str(uid)
    env['HOME'] = home
    env['XAUTHORITY'] = join(home, '.Xauthority')
    env['ICEAUTHORITY'] = join(home, '.ICEauthority')
    env['SUGAR_ACTIVITY_ROOT'] = home
    env['TMPDIR'] = join(home, 'tmp')

    # Check for use-camera, and use-micropphone
    # Add appropriate groups from /etc/group
    groups = []
    gset = False
    if pset != None:
        if pset.has_permission('use-camera'):
            groups.append(grp.getgrnam('camera')[2])
            log ('Turning on camera access')
            gset = True

        if pset.has_permission('use-microphone'):
            log ('Turning on microphone access')
            groups.append(grp.getgrnam('microp')[2])
            gset = True

    #
    # Set appropriate group membership(s), depending on requested permissions
    # List of memberships in 'groups'
    #
    if gset == True:
        os.setgroups(groups)

    log('dropping privilege to (%d, %d)', uid, gid)
    os.setgid(gid)
    os.setuid(uid)

    # Limit various resources
    # Must be done *after* setting uid/gid
    # This should come from the permissions.info file, but this is OK
    #   for now.
    try:
        def set_limit(lim_name, unix_name):
            p = pset.permission_params('lim_' + lim_name)
            if p != None:
                x = int(float(p[0]))
                y = int(float(x*1.10))
                log('Setting RLIMIT_%s to %d,%d', unix_name, x, y)
                resource.setrlimit(getattr(resource, 'RLIMIT_'+unix_name), (x,y))

        set_limit('nofile', 'NOFILE')
        set_limit('fsize', 'FSIZE')
        set_limit('mem', 'AS')
        set_limit('nproc', 'NPROC')

    except:
        pass # Why would we be throwing exceptions when setting rlimits? <MS>

    log('chdir to %s' % cwd)
    os.chdir(cwd)

    log('umask(0)')
    os.umask(0)

    if strace_hint:
        argv, env = strace(log, argv, env)

    log('about to execve\nargv: %s\nenv: %s', argv, env)
    log('closing all fds but %s', safe_fds)
    for fd in get_fds():
        if fd not in safe_fds:
            try: os.close(fd) # propagate failure from EIO or EBADF.
            except: pass

    # This looks a bit hacky. Probably need a better way to know that this
    # activitiy needs to go through the fast path.
    if preloader_hint and argv[0] == 'sugar-activity':

        # hack: should be parameters to sugarlaunch.launch()
        sys.argv = argv

        for key, value in env.iteritems():
            os.environ[key] = value

        log(os.environ)

        # Recreate the connection to X
        import gtk
        d = gtk.gdk.Display(':0')
        dm = gtk.gdk.display_manager_get()
        dm.set_default_display(d)

        from rainbow import sugaractivity
        sugaractivity.launch()
    else:
        os.execvpe(argv[0], argv, env)

def check_bundle_id(bundle_id):
    assert bundle_id and all(s.isalnum() for s in bundle_id.split('.'))

def check_bundle_path(uid, gid, path):
    ck = Checker(path, uid, gid)
    assert ck.positive(F_OK | R_OK, 0) and isabs(path)

def check_argv(argv):
    assert argv and len(argv) > 1

def check_cwd(uid, gid, cwd):
    ck = Checker(cwd, uid, gid)
    assert ck.positive(R_OK | X_OK, S_IFDIR)
    assert ck.negative(W_OK, 0)

def check_spool(spool, owner_uid, owner_gid):
    spool_dirs = ('uid_pool', 'gid_pool', 'bundle_id_to_gid',
                 'uid_to_instance_dir', 'gid_to_data_dir', 'uid_to_home_dir',
                 'uid_to_gid')
    for frag in spool_dirs:
        make_dirs(join(spool, frag), owner_uid, owner_gid, 0775)
        ck = Checker(join(spool, frag), owner_uid, owner_gid)
        assert ck.positive(R_OK | W_OK | X_OK, S_IFDIR)

def check_owner(owner_uid, owner_gid):
    assert (owner_uid, owner_gid) == (500, 500)

def check_home_dirs(uid, gid, home):
    for frag in ('instance', 'data', 'tmp'):
        ck = Checker(join(home, frag), uid, gid)
        assert ck.positive(R_OK | W_OK | X_OK, S_IFDIR)

def check_home(uid, gid, home):
    ck = Checker(home, uid, gid)
    assert ck.positive(R_OK | W_OK | X_OK, S_IFDIR)

def run(log, spool, env, argv, cwd, pset, safe_fds, strace_hint,
        owner_uid, owner_gid, bundle_path, bundle_id, constant_uid,
        preloader_hint):
    # Note: exceptions are intended to bubble up to the caller and should
    # terminate execution.
    check_bundle_id(bundle_id)
    check_argv(argv)
    check_owner(owner_uid, owner_gid)

    check_spool(spool, owner_uid, owner_gid)

    uid, gid = reserve_credentials(log, spool, bundle_id, constant_uid)
    home = grab_home(log, spool, bundle_id, uid, gid, owner_gid)
    configure_home(spool, owner_uid, owner_gid, uid, gid, home)

    if cwd is None:
        cwd = home
    check_cwd(uid, gid, cwd)
    check_home_dirs(uid, gid, home)
    check_home_dirs(owner_uid, owner_gid, home)
    check_home(uid, gid, home)
    check_bundle_path(uid, gid, bundle_path)

    launch(log, home, uid, gid, argv, env, cwd, pset, safe_fds, strace_hint, preloader_hint)
